# cc2-gr4.py
# L1-Info, Nice, Automne 2016

print("---- Definition d'un polynome dense ----")

p1 = [-2,0,6,-3]
print('p1 =',p1,' soit -2+6x**2-3x**3')

def degre(p) :     # p est un polydense
    return len(p) - 1

print('p1 est un polynome dense de degre',degre(p1))
print()

print("---- Valeur d'un polynome en un point ----")

def valeur(p,x) :
    return sum(p[i]*x**i for i in range(len(p)))    # quick and dirty

def valeur(p,x) :                 # schema de Horner
    res = 0
    for i in range(1,len(p)+1) :   # parcours droite -> gauche de p
        res = res * x + p[-i]     
    return res

print('Valeur de p1 en x=-1 :',valeur(p1,-1))    # 7
print()

print("---- Derivation d'un polynome ----")

def derive(p) :
    return [i*p[i] for i in range(1,len(p))]

dp1 = derive(p1)
print("Le polynome derive dp1 est",dp1,'de degre',degre(dp1))   # [0,-8,6]
print()

print('---- Dichotomie lorsque p est monotone sur [a,b] ----')

def dichotomie(p,a,b,h) :    # p est strictement monotone sur [a,b]
    ya = valeur(p,a) ; yb = valeur(p,b)
    if ya * yb < 0 :    # il y a une seule racine dans ]a,b[. Cherchons-la.
        while True :    # recherche par dichotomie ! Voir hoering.py
            m = (a + b) / 2
            if abs(valeur(p,m)) < h : return m
            if valeur(p,a) * valeur(p,m) < 0 : b = m
            else : a = m
    else : return False

print('Je travaille par dichotomie sur le polynome p1 = -2+6*x**2-3x**3, dans')
print('des intervalles ou il est strictement monotone. Voir la figure 1.')
print("Recherche d'une racine dans [-1,0] :",end=' ')
r = dichotomie(p1,-1,0,0.01)
print('r =',r)      # r = -0.515625
print("Recherche d'une racine dans [0,1] :",end=' ')
r = dichotomie(p1,0,1,0.01)
print('r =',r)      # r = 0.72265625
print("Recherche d'une racine dans [1.5,2] :",end=' ')
r = dichotomie(p1,1.5,2,0.01)
print('r =',r)      # r = r = 1.79296875
print("Recherche d'une racine dans [-2,-0.7] :",end=' ')
r = dichotomie(p1,-2,-0.7,0.01)
print('r =',r)      # r = False
print()

print('---- Le cas du degre 1 (non utilise par la suite) ---')

def racine_degre1(p,a,b) :    # p est de degre 1
    [c0,c1] = p               # et on sait que c1 != 0
    racine = -c0/c1
    if a <= racine <= b : return racine
    return False

print("Je travaille sur le polynome 3x-2 de degre 1.")
print("Dans l'intervalle [-5,5] la racine est :",racine_degre1([-2,3],-5,5))
print("Dans l'intervalle [1,5] la racine est :",racine_degre1([-2,3],1,5))
print()

print("---- Les extrema d'un polynome dans [a,b] ---")

# Etape 1 : calcul de la liste des extrema de p dans [a,b]. On compte les extremites
# a et b comme extrema. On suppose la fonction racines deja ecrite ! Les
# sommets d'un polynome sont obtenus en annullant la derivee premiere.

def extrema(p,a,b,h) :        # on suppose p de degre >= 1
    if degre(p) < 1 : return []   # oups, p est constant et non nul !
    return [a] + racines(derive(p),a,b,h) + [b]

# Etape 2 : on programme la fonction racines qui va utiliser la fonction extrema.
# Elle retourne la liste des racines simples de p dans [a,b]. Pour cela, elle va
# detecter les sous-intervalles ou p est strictement monotone.

def racines(p,a,b,h) :          # on suppose p de degre >= 1
    # si p est constant non nul, pas de racine !
    if degre(p) == 0 :  return []
    # sinon on decoupe [a,b] par les extrema et on travaille entre les extrema !
    L = extrema(p,a,b,h)
    res = []
    for i in range(len(L)-1) :
        #print('Je travaille entre',L[i],'et',L[i+1])
        essai = dichotomie(p,L[i],L[i+1],h)  # tentative entre deux extrema consecutifs
        if essai : res.append(essai)
    return res

p2 = [-150,35,-121,24,30,-11,1]
print('Je travaille sur le polynome p2 = [-150,35,-121,24,30,-11,1] de degre 6, qui')
print('provient du developpement de (x+2)(x-3)(x**2+1)(x-5)**2. Il ne possede que')
print('deux racines simples reelles -2 et 3. Voir la figure 2.')
print('-Extrema de p2 dans [-4,6] en calculant les racines de sa derivee :')
print('  ',extrema(p2,-4,6,0.01))
print('-racines simples de p2 dans [-10,10] :')
print('  ',racines(p2,-10,10,0.001))
print('-racines simples de p2 dans [4,10] :')
print('  ',racines(p2,4,10,0.001),'   (racine double !)')

print("Je travaille avec le polynome p3 qui s'ecrit -3+x**7, pour calculer")
print('la racine septieme de 3 qui vaut :',end=' ')
p3 = [-3,0,0,0,0,0,0,1]
print(racines(p3,0,10,0.001)[0])

