# Representation creuse des polynomes en Python. Solutions de la seance 7.
# Faculte des Sciences de Nice, niveau L1.
# Un polynome est une liste [[c,e],...] de monomes suivant les puissances decroissantes
# Un monome est une liste [c,e] dans lequel le coeff c est != 0, et e >= 0 est entier
# poly0 est une fonction !
#
# polycreux.py      alias      sol7.py
#
# la variable __all__ si elle est definie, limite les noms exportes par un 'from polycreux import *'
__all__ = ('poly0', 'degre', 'coef', 'add', 'mul_ext', 'sub', 'mul',
           'fpoly', 'deriv', 'dessiner')

print('Module polycreux : poly0(), degre(mp), coeff(mp), add(p,q), sub(p,q), mul_ext(k,p),')
print('mul(p,q), fpoly(p), deriv(p), dessiner(p,a,u=1,v=1).\n')

# Mettez True pour executer tous les tests ! <--------------------------------------
polycreux_exec_test = False
if polycreux_exec_test : print("***** Tests du module 'polycreux' *****")

########################### le module polycreux ####################################

def poly0() :          # le constructeur du polynome nul
    return []

def is_poly0(p) : 
    return p == []

def degre(mp) :        # degre d'un monome [c,e] ou d'un polynome [[c,e],...]
    if is_poly0(mp) : return float("-Inf") 
    if isinstance(mp[0],list) : return mp[0][1]
    return mp[1]

def coef(mp) :         # coefficient dominant d'un monome [c,e] ou d'un polynome [[c,e],...]
    if is_poly0(mp) : return 0
    if isinstance(mp[0],list) : return mp[0][0]
    return mp[0]

if polycreux_exec_test :          # deux exemples de polynomes pour les tests
    global p                      # pour que p soit defini au toplevel !
    p = [[2,5],[-1,4],[-2,1]]
    print('p =',p)
    global q                      # pour que q soit defini au toplevel !
    q = [[1,4],[7,2],[-1,0]]
    print('q =',q)
    print('Il nous manque une fonction affiche(p) qui retournerait')
    print("la string '2x^5-x^4-2'. Un(e) volontaire ?...") 

def mono_plus_poly(c,e,p) :   # le monome cx^e ajoute (a sa place) au polynome p. Iteratif !
    # algo d'insertion : on cherche le premier monome de degré d <= e
    for i in range(len(p)) :
        [ci,ei] = p[i]        # le monome numero i de p : mieux que m = p[i]
        if ei == e : 
            essai = c + ci
            if essai == 0 : return p[:i] + p[i+1:]
            return p[:i] + [[essai,e]] + p[i+1:]
        if ei < e :           # insertion d'un nouvel element !
            return p[:i] + [[c,e]] + p[i:]
    return p + [[c,e]]

if polycreux_exec_test :      # test de mono_plus_poly
    print('x^4 + p =',mono_plus_poly(1,4,p))
    print('5x^3 + p =',mono_plus_poly(5,3,p))
    print('1 + p =',mono_plus_poly(1,0,p))

def add(p1,p2) :
    for m in p1 :
        p2 = mono_plus_poly(coef(m),degre(m),p2)
    return p2

if polycreux_exec_test :      # test de add
    print('p + q =',add(p,q))

def mul_ext(k,p) :            # loi externe : nombre x polynome --> polynome
    res = poly0()
    for m in p :
        res.append([coef(m) * k, degre(m)])
    return res

if polycreux_exec_test :      # test de mul_ext
    print('2p =',mul_ext(2,p))

def mul_ext_v2(k,p) :         # comprehension de liste
    if k == 0 : return poly0()
    return [[coef(m) * k, degre(m)] for m in p]

if polycreux_exec_test :      # test de mul_ext_v2
    print('2p =',mul_ext_v2(2,p))

def mul_ext_v3(k,p) :         # comprehension de liste + destructuration
    if k == 0 : return poly0()
    return [[c*k, e] for [c,e] in p]      # [c,e] au lieu du maladroit m

if polycreux_exec_test :      # test de mul_ext_v3
    print('2p =',mul_ext_v3(2,p))

def sub(p1,p2) :              # soustraction de deux polynomes
    return add(p1,mul_ext(-1,p2))

if polycreux_exec_test :      # test de sub
    print('p - q =',sub(p,q))
    print('p - p =',sub(p,p))

def mono_mul_poly(c,e,p) : 
    res = poly0()
    for [cp,ep] in p :        # soit [cp,ep] le monome courant de p
        res = mono_plus_poly(c * cp, e + ep,res)
    return res

if polycreux_exec_test :      # test de mono_mul_poly
    print('2x^3 * p =',mono_mul_poly(2,3,p))

def mono_mul_poly_v2(c,e,p) : 
    res = poly0()
    for [ci,ei] in p : 
        res.append([c * ci, e + ei])
    return res

if polycreux_exec_test :      # test de mono_mul_poly
    print('2x^3 * p =',mono_mul_poly_v2(2,3,p))

def mul(p,q) :   # multiplication de deux polynomes. Iteratif ! Avec simple boucle
    res = poly0()
    for [ci,ei] in p :
        res = add(res,mono_mul_poly(ci,ei,q))
    return res

if polycreux_exec_test :      # test de mul
    print('p * q =',mul(p,q))
    print('(x^100-1)*(x^100+1) =',mul([[1,100],[-1,0]],[[1,100],[1,0]]))

def valeur(p,x) :             # valeur du polynome p au point x
    res = 0
    for [c,e] in p :
        res = res + c*(x**e)
    return res

if polycreux_exec_test :      # test de valeur
    print('Valeur de p en 2 :',valeur(p,2))   # 44
    if valeur(p,2) != 44 : raise ValueError('Bug dans la fonction valeur version 1 !')

def valeur_v2(p,x) :
    return sum(c*(x**e) for [c,e] in p) 

if polycreux_exec_test :      # test de valeur
    if valeur_v2(p,2) != 44 : raise ValueError('Bug dans la fonction valeur_v2 !')

def fpoly(p) :                # la fonction polynome associee au polynome p
    def g(x) :
        return valeur(p,x)
    return g

def fpoly_v2(p) :
    return lambda x : valeur(p,x)     # lambda-fonction == fonction sans nom

if polycreux_exec_test :      # test de fpoly
    global fp                 # pour que fp soit visible au toplevel lors du graphisme
    fp = fpoly(p)
    print('Soit fp la fonction associee au polynome p.')
    print('fp(1) =',fp(1))

################### TP 7 ###################

def deriv(p) :                # le polynome derive p' 
    res = poly0()
    for [c,e] in p :
        if e != 0 : res.append([c*e,e-1]) 
    return res

if polycreux_exec_test :      # test de deriv
    print('Avec une boucle : deriv(q) =',deriv(q))

def deriv_v2(p) :
    return [[c*e,e-1] for [c,e] in p if e != 0]    # comprehension + destructuration

if polycreux_exec_test :      # test de deriv_v2
    print('En pur style Python : deriv(q) =',deriv(q))

################# TurtleGraphics == Fun time :-) #############################

from turtle import *

def texte(str,x,y) :     # on peut ecrire dans la fenetre graphique...
    up() ; goto(x,y) ; write(str,font=('Times',20,'normal'))

def point(x,y) :
    up() ; goto(x,y); down(); dot(5)
        
def axes(u,v) :     # dessin des axes + point sur les vecteurs unitaires de longueurs u et v
    p = pencolor()  # sauvegarde de la couleur courante
    pencolor("black")
    up() ; goto(-300,0); down() ;  goto(300,0)
    up() ; goto(0,300); down() ; goto(0,-300)
    texte('x',260,5) ; texte('y',10,265)
    point(0,0); point(u,0) ; point(0,v)
    pencolor(p)

# ATTENTION : ne pas confondre les x mathématiques et les x en pixels ! En fait :
#     x_pixels = x_math * u     et     y_pixels = y_math * v
# si u et v sont les longueurs en pixels des vecteurs unitaires !
# On dessine sur [-a,a] ou a est en unite mathematique. Par exemple dans l'intervalle [-2,2],
# ce qui ne signifie bien entendu pas entre -2 pixels et 2 pixels !!! C'est la seule difficulte
# de cet exercice si vous avez assimile le cours 4 page 7...

def dessiner(p,a,u=1,v=1) :  # u et v = longueurs des vecteurs unitaires, dessin pour x in [-a,a]
    f = fpoly(p)             # mieux vaut travailler avec une vraie fonction !
    axes(u,v)
    x = -a ; y = f(x)        # x et y sont des unites mathematiques
    up() ; goto(x*u,y*v) ; down()
    while x < a : 
        xs = x + 0.1 ; ys = f(xs)
        goto(xs*u,ys*v)
        (x,y) = (xs,ys)
                
# Etudiez bien les correlations entre le signe de la fonction en rouge et les
# variations de la fonction en bleu !

if polycreux_exec_test :      # test de dessiner
    p = [[1,3],[-5,1]]           # y = x^3-5x
    ht()                         # pour ne pas voir le chelonien...
    tracer(False)                # pour dessiner tres vite
    pencolor("blue")
    dessiner(p,2,120,40)
    texte('p',230,-140)
    pencolor("red")
    dessiner(deriv(p),2,120,40)   # y = 3x^2-5
    texte("p'",230,160)
    pencolor("black")
    texte("p '(x) = 0",-220,-175)
    up() ; goto(-200,-150) ; down(); goto(-155,-5)
    texte(' sommet pour p',-150,250)
    up() ; goto(-130,250) ; down(); goto(-155,175)
    texte('Regardez les comportements de ces deux courbes en termes de',-280,-240)
    texte('la fonction f et ses d\u00e9riv\u00e9es premi\u00e8re et seconde...',-280,-265)
    tracer(True)

if polycreux_exec_test == True :
    print('\nMettez False au lieu de True comme valeur de la variable polycreux_exec_test')
    print('en haut de ce fichier (ligne 17) pour masquer les tests de polycreux.py !')






